@@ -10,7 +10,7 @@ from maitai._config import config
 from maitai._maitai_client import MaitaiClient
 from maitai._utils import __version__ as version, chat_completion_chunk_to_response
 from maitai_common.utils.types import EvaluateCallback
-from maitai_gen.chat import ChatCompletionChunk, ChatCompletionParams, ChatCompletionResponse, EvaluateRequest, EvaluateResponse, EvaluationContentType
+from maitai_gen.chat import ChatCompletionChunk, ChatCompletionParams, ChatCompletionRequest, ChatCompletionResponse, EvaluateRequest, EvaluateResponse, EvaluationContentType
 from maitai_gen.metric import RequestTimingMetric
 
 
@@ -30,7 +30,7 @@ class Evaluator(MaitaiClient):
 
     @classmethod
     async def evaluate_async(cls, session_id, reference_id, intent, content, content_type=None, application_id=None, application_ref_name=None, callback=None, completion_params: ChatCompletionParams = None,
-                             chat_completion_response: ChatCompletionResponse = None, chat_completion_chunk: ChatCompletionChunk = None, timing: RequestTimingMetric = None):
+                             chat_completion_response: ChatCompletionResponse = None, chat_completion_chunk: ChatCompletionChunk = None, timing: RequestTimingMetric = None, metadata: dict = {}):
         if content_type is None:
             content_type = _get_content_type(content)
         if content_type is None:
@@ -39,7 +39,18 @@ class Evaluator(MaitaiClient):
             raise Exception('application_id or application_ref_name must be provided')
         eval_request: EvaluateRequest = cls.create_eval_request(application_id, application_ref_name, session_id, reference_id, intent, content_type, content)
         if completion_params is not None:
-            eval_request.chat_completion_request = cls.create_inference_request(application_ref_name, session_id, reference_id, intent, False, True, completion_params, callback, False, '')
+            eval_request.chat_completion_request = ChatCompletionRequest(
+                application_ref_name=application_ref_name,
+                session_id=session_id,
+                reference_id=reference_id,
+                action_type=intent,
+                apply_corrections=False,
+                evaluation_enabled=True,
+                params=completion_params,
+                return_evaluation=True if callback else False,
+                auth_keys=config.auth_keys,
+                metadata=metadata,
+            )
             if chat_completion_chunk is not None:
                 chat_completion_response = chat_completion_chunk_to_response(chat_completion_chunk, content)
             eval_request.chat_completion_response = chat_completion_response
@@ -48,7 +59,7 @@ class Evaluator(MaitaiClient):
 
     @classmethod
     def evaluate(cls, session_id, reference_id, intent, content, content_type=None, application_id=None, application_ref_name=None, callback=None, partial=None, completion_params: ChatCompletionParams = None,
-                 chat_completion_response: ChatCompletionResponse = None, chat_completion_chunk: ChatCompletionChunk = None, timing: RequestTimingMetric = None):
+                 chat_completion_response: ChatCompletionResponse = None, chat_completion_chunk: ChatCompletionChunk = None, timing: RequestTimingMetric = None, metadata: dict = {}):
         if content_type is None:
             content_type = _get_content_type(content, partial)
         if content_type is None:
@@ -57,7 +68,18 @@ class Evaluator(MaitaiClient):
             raise Exception('application_id or application_ref_name must be provided')
         eval_request: EvaluateRequest = cls.create_eval_request(application_id, application_ref_name, session_id, reference_id, intent, content_type, content, partial=partial)
         if completion_params is not None:
-            eval_request.chat_completion_request = cls.create_inference_request(application_ref_name, session_id, reference_id, intent, False, True, completion_params, callback, False, None)
+            eval_request.chat_completion_request = ChatCompletionRequest(
+                application_ref_name=application_ref_name,
+                session_id=session_id,
+                reference_id=reference_id,
+                action_type=intent,
+                apply_corrections=False,
+                evaluation_enabled=True,
+                params=completion_params,
+                return_evaluation=True if callback else False,
+                auth_keys=config.auth_keys,
+                metadata=metadata,
+            )
             if chat_completion_chunk is not None:
                 chat_completion_response = chat_completion_chunk_to_response(chat_completion_chunk, content)
             eval_request.chat_completion_response = chat_completion_response
@@ -97,23 +119,29 @@ class Evaluator(MaitaiClient):
 
     @classmethod
     async def send_evaluation_request_async(cls, eval_request: EvaluateRequest, callback: EvaluateCallback = None):
+        path = "request" if callback else "submit"
+        host = config.maitai_host
+        url = f'{host}/evaluation/{path}'
+
         async def send_request():
-            path = "request" if callback else "submit"
-            host = config.maitai_host
-            url = f'{host}/evaluation/{path}'
             headers = {
                 'Content-Type': 'application/json',
                 'x-api-key': config.api_key,
                 'x-client-version': version,
             }
-            async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=False)) as session:
-                async with session.post(url, headers=headers,
-                                        data=eval_request.to_json(casing=Casing.SNAKE)) as response:
-                    if response.status != 200:
-                        error_text = await response.text()
-                        print(f"Failed to send evaluation request. Status code: {response.status}. Error: {error_text}")
-                        return None
-                    return await response.read()
+            try:
+                async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=False)) as session:
+                    async with session.post(url, headers=headers,
+                                            data=eval_request.to_json(casing=Casing.SNAKE)) as response:
+                        print("received", response.status)
+                        if response.status != 200:
+                            error_text = await response.text()
+                            cls.log_error(error_text, url)
+                            print(f"Failed to send evaluation request. Status code: {response.status}. Error: {error_text}")
+                            return None
+                        return await response.read()
+            except Exception as e:
+                await cls.increment_error(str(e), url)
 
         result = await send_request()
         if result is not None:
@@ -140,6 +168,7 @@ class Evaluator(MaitaiClient):
             response = requests.post(url, headers=headers, data=eval_request.to_json(casing=Casing.SNAKE), stream=True)
             if response.status_code != 200:
                 print(f"Failed to send stream correction request. Status code: {response.status_code}. Error: {response.text}")
+                cls.log_error(response.text, url)
                 return
             try:
                 for line in response.iter_lines():
