@@ -97,11 +97,11 @@ class Completions:
         self._client = client
         self._client_params = client_params or ClientParams()
 
-    @required_args(["action_type", "application_ref_name", "messages"],
+    @required_args(["intent", "application", "messages"],
                    ["intent", "application_ref_name", "messages"],
-                   ["action_type", "application_ref_name", "messages", "model"],
+                   ["intent", "application", "messages", "model"],
                    ["intent", "application_ref_name", "messages", "model"],
-                   ["action_type", "application_ref_name", "messages", "model", "stream"],
+                   ["intent", "application", "messages", "model", "stream"],
                    ["intent", "application_ref_name", "messages", "model", "stream"])
     def create(
         self,
@@ -112,7 +112,8 @@ class Completions:
         reference_id: Union[str, int, None] = None,
         intent: str = None,
         action_type: str = None,  # DEPRECATED
-        application_ref_name: str = None,
+        application_ref_name: str = None, # DEPRECATED
+        application: str = None,
         callback: Optional[EvaluateCallback] = None,
         server_side_inference: bool = None,
         evaluation_enabled: bool = None,
@@ -122,6 +123,7 @@ class Completions:
         return_request: bool = False,
         fallback_model: str = None,
         assistant: bool = False,
+        metadata: dict = {},
         # OpenAI Arguments
         messages: Iterable[openai_chat_types.ChatCompletionMessageParam],
         model: Union[str, openai_types.ChatModel, openai.NotGiven] = openai.NOT_GIVEN,
@@ -148,6 +150,7 @@ class Completions:
         extra_query: Optional[Query] = None,
         extra_body: Optional[Body] = None,
         timeout: Union[float, httpx.Timeout, None, openai.NotGiven] = openai.NOT_GIVEN,
+        parallel_tool_calls: Union[bool, openai.NotGiven] = openai.NOT_GIVEN,
     ) -> Union[MaitaiCompletion, Iterable[MaitaiChunk]]:
         retry_args = locals()
         timing = RequestTimingMetric(
@@ -159,10 +162,19 @@ class Completions:
             raise ValueError("server_side_inference must be true to apply_corrections")
         if apply_corrections is True and evaluation_enabled is False:
             raise ValueError("evaluations must be enabled to apply_corrections")
+        if assistant is True:
+            user_msgs = [True for msg in messages if msg["role"] == "user"]
+            assistant_messages = [True for msg in messages if msg["role"] == "assistant"]
+            if len(user_msgs) != 1:
+                raise ValueError("When assistant = True, exactly 1 user message is required")
+            if len(assistant_messages) > 0:
+                raise ValueError("When assistant = True, no assistant messages are allowed")
         if not session_id:
             session_id = str(uuid.uuid4())
         if not intent and action_type:
             intent = action_type
+        if not application_ref_name and application:
+            application_ref_name = application
         maitai_config = config.get_application_action_config(application_ref_name, intent)
         if server_side_inference is None:
             server_side_inference = maitai_config.inference_location == InferenceLocations.SERVER
@@ -218,28 +230,32 @@ class Completions:
             top_logprobs=top_logprobs,
             top_p=top_p,
             user=user,
+            parallel_tool_calls=parallel_tool_calls,
         )
         if server_side_inference:
             response_timeout = None
             if isinstance(timeout, float) or isinstance(timeout, int):
                 response_timeout = timeout
-            chat_request: ChatCompletionRequest = maitai.Inference.create_inference_request(
+            chat_request: ChatCompletionRequest = ChatCompletionRequest(
                 application_ref_name=application_ref_name,
                 session_id=session_id,
                 reference_id=reference_id,
-                intent=intent,
+                action_type=intent,
                 apply_corrections=apply_corrections,
                 evaluation_enabled=evaluation_enabled,
-                completion_params=completion_params,
-                callback=callback,
+                params=completion_params,
+                return_evaluation=True if callback else False,
                 context_retrieval_enabled=context_retrieval_enabled,
                 context_query=context_query,
                 return_request=return_request,
                 fallback_model=fallback_model,
                 user_id=user_id,
                 assistant=assistant,
-                client_params=self._client_params,
+                auth_keys=config.auth_keys,
+                metadata=metadata,
             )
+            if self._client_params is not None:
+                chat_request.client_params = self._client_params
             try:
                 timing.time_pre_request = time.time()
                 response = maitai.Inference.infer(chat_request, callback, response_timeout)
@@ -301,10 +317,11 @@ class Completions:
                 extra_query=extra_query,
                 extra_body=extra_body,
                 timeout=timeout,
+                parallel_tool_calls=parallel_tool_calls,
 
             )
             if stream:
-                return _process_openai_stream(session_id, reference_id, intent, application_ref_name, messages, response, evaluation_enabled, completion_params, timing, callback)
+                return _process_openai_stream(session_id, reference_id, intent, application_ref_name, messages, response, evaluation_enabled, completion_params, timing, callback, metadata)
             else:
                 timing.time_request_end = time.time()
                 maitai_completion = convert_openai_chat_completion(response)
@@ -323,6 +340,7 @@ class Completions:
                         chat_completion_response=maitai_completion,
                         completion_params=completion_params,
                         timing=timing,
+                        metadata=metadata,
                     )
                 else:
                     maitai.Inference.store_chat_response(
@@ -335,6 +353,7 @@ class Completions:
                         final_chunk=None,
                         content="",
                         timing=timing,
+                        metadata=metadata,
                     )
                 return MaitaiCompletion(maitai_completion)
 
@@ -349,6 +368,7 @@ def _process_openai_stream(session_id: Union[str, int],
                            chat_completion_params: ChatCompletionParams,
                            timing: RequestTimingMetric,
                            callback: Optional[EvaluateCallback] = None,
+                           metadata: dict = {},
                            ) -> Iterable[MaitaiChunk]:
     full_body = ""
     proto_messages = openai_messages_to_proto(messages)
@@ -379,6 +399,7 @@ def _process_openai_stream(session_id: Union[str, int],
             chat_completion_chunk=last_chunk,
             completion_params=chat_completion_params,
             timing=timing,
+            metadata=metadata,
         )
     else:
         maitai.Inference.store_chat_response(
@@ -391,6 +412,7 @@ def _process_openai_stream(session_id: Union[str, int],
             content=full_body,
             chat_completion_response=None,
             timing=timing,
+            metadata=metadata,
         )
 
 
