@@ -7,7 +7,7 @@ def get_default_config() -> Config:
     return Config(
         inference_location=InferenceLocations.SERVER,
         evaluation_enabled=True,
-        apply_corrections=True,
+        apply_corrections=False,
         model='gpt-4o',
         temperature=1,
         streaming=False,
@@ -18,7 +18,7 @@ def get_default_config() -> Config:
         n=1,
         frequency_penalty=0,
         presence_penalty=0,
-        timeout=-1,
+        timeout=0,
         context_retrieval_enabled=False,
         fallback_model=None,
     )
